# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""

import grpc

import tekhsi._tek_highspeed_server_pb2 as TekHighspeedServer__pb2


class ConnectStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel: grpc.Channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Connect = channel.unary_unary(
            "/Tekscope.Connect/Connect",
            request_serializer=TekHighspeedServer__pb2.ConnectRequest.SerializeToString,
            response_deserializer=TekHighspeedServer__pb2.ConnectReply.FromString,
        )
        self.Disconnect = channel.unary_unary(
            "/Tekscope.Connect/Disconnect",
            request_serializer=TekHighspeedServer__pb2.ConnectRequest.SerializeToString,
            response_deserializer=TekHighspeedServer__pb2.ConnectReply.FromString,
        )
        self.RequestNewSequence = channel.unary_unary(
            "/Tekscope.Connect/RequestNewSequence",
            request_serializer=TekHighspeedServer__pb2.ConnectRequest.SerializeToString,
            response_deserializer=TekHighspeedServer__pb2.ConnectReply.FromString,
        )
        self.RequestAvailableNames = channel.unary_unary(
            "/Tekscope.Connect/RequestAvailableNames",
            request_serializer=TekHighspeedServer__pb2.ConnectRequest.SerializeToString,
            response_deserializer=TekHighspeedServer__pb2.AvailableNamesReply.FromString,
        )
        self.WaitForDataAccess = channel.unary_unary(
            "/Tekscope.Connect/WaitForDataAccess",
            request_serializer=TekHighspeedServer__pb2.ConnectRequest.SerializeToString,
            response_deserializer=TekHighspeedServer__pb2.ConnectReply.FromString,
        )
        self.FinishedWithDataAccess = channel.unary_unary(
            "/Tekscope.Connect/FinishedWithDataAccess",
            request_serializer=TekHighspeedServer__pb2.ConnectRequest.SerializeToString,
            response_deserializer=TekHighspeedServer__pb2.ConnectReply.FromString,
        )


class ConnectServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Connect(self, request, context):
        """Connect to instrument."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def Disconnect(self, request, context):
        """Disconnect from instrument."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def RequestNewSequence(self, request, context):
        """Force new sequence.

        This requests access to data.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def RequestAvailableNames(self, request, context):
        """Returns a list of names of available data."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def WaitForDataAccess(self, request, context):
        """Waits for client access to data.

        When it returns you may access NormalizedData or NativeData.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def FinishedWithDataAccess(self, request, context):
        """Releases access to data (must occur after WaitForDataAccess)"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_ConnectServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "Connect": grpc.unary_unary_rpc_method_handler(
            servicer.Connect,
            request_deserializer=TekHighspeedServer__pb2.ConnectRequest.FromString,
            response_serializer=TekHighspeedServer__pb2.ConnectReply.SerializeToString,
        ),
        "Disconnect": grpc.unary_unary_rpc_method_handler(
            servicer.Disconnect,
            request_deserializer=TekHighspeedServer__pb2.ConnectRequest.FromString,
            response_serializer=TekHighspeedServer__pb2.ConnectReply.SerializeToString,
        ),
        "RequestNewSequence": grpc.unary_unary_rpc_method_handler(
            servicer.RequestNewSequence,
            request_deserializer=TekHighspeedServer__pb2.ConnectRequest.FromString,
            response_serializer=TekHighspeedServer__pb2.ConnectReply.SerializeToString,
        ),
        "RequestAvailableNames": grpc.unary_unary_rpc_method_handler(
            servicer.RequestAvailableNames,
            request_deserializer=TekHighspeedServer__pb2.ConnectRequest.FromString,
            response_serializer=TekHighspeedServer__pb2.AvailableNamesReply.SerializeToString,
        ),
        "WaitForDataAccess": grpc.unary_unary_rpc_method_handler(
            servicer.WaitForDataAccess,
            request_deserializer=TekHighspeedServer__pb2.ConnectRequest.FromString,
            response_serializer=TekHighspeedServer__pb2.ConnectReply.SerializeToString,
        ),
        "FinishedWithDataAccess": grpc.unary_unary_rpc_method_handler(
            servicer.FinishedWithDataAccess,
            request_deserializer=TekHighspeedServer__pb2.ConnectRequest.FromString,
            response_serializer=TekHighspeedServer__pb2.ConnectReply.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler("Tekscope.Connect", rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class Connect(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Connect(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/Tekscope.Connect/Connect",
            TekHighspeedServer__pb2.ConnectRequest.SerializeToString,
            TekHighspeedServer__pb2.ConnectReply.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def Disconnect(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/Tekscope.Connect/Disconnect",
            TekHighspeedServer__pb2.ConnectRequest.SerializeToString,
            TekHighspeedServer__pb2.ConnectReply.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def RequestNewSequence(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/Tekscope.Connect/RequestNewSequence",
            TekHighspeedServer__pb2.ConnectRequest.SerializeToString,
            TekHighspeedServer__pb2.ConnectReply.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def RequestAvailableNames(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/Tekscope.Connect/RequestAvailableNames",
            TekHighspeedServer__pb2.ConnectRequest.SerializeToString,
            TekHighspeedServer__pb2.AvailableNamesReply.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def WaitForDataAccess(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/Tekscope.Connect/WaitForDataAccess",
            TekHighspeedServer__pb2.ConnectRequest.SerializeToString,
            TekHighspeedServer__pb2.ConnectReply.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def FinishedWithDataAccess(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/Tekscope.Connect/FinishedWithDataAccess",
            TekHighspeedServer__pb2.ConnectRequest.SerializeToString,
            TekHighspeedServer__pb2.ConnectReply.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )


class NormalizedDataStub(object):
    """Normalized Waveform (this is fully defined)"""

    def __init__(self, channel: grpc.Channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetWaveform = channel.unary_stream(
            "/Tekscope.NormalizedData/GetWaveform",
            request_serializer=TekHighspeedServer__pb2.WaveformRequest.SerializeToString,
            response_deserializer=TekHighspeedServer__pb2.NormalizedReply.FromString,
        )
        self.GetHeader = channel.unary_unary(
            "/Tekscope.NormalizedData/GetHeader",
            request_serializer=TekHighspeedServer__pb2.WaveformRequest.SerializeToString,
            response_deserializer=TekHighspeedServer__pb2.NormalizedReply.FromString,
        )


class NormalizedDataServicer(object):
    """Normalized Waveform (this is fully defined)"""

    def GetWaveform(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetHeader(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_NormalizedDataServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "GetWaveform": grpc.unary_stream_rpc_method_handler(
            servicer.GetWaveform,
            request_deserializer=TekHighspeedServer__pb2.WaveformRequest.FromString,
            response_serializer=TekHighspeedServer__pb2.NormalizedReply.SerializeToString,
        ),
        "GetHeader": grpc.unary_unary_rpc_method_handler(
            servicer.GetHeader,
            request_deserializer=TekHighspeedServer__pb2.WaveformRequest.FromString,
            response_serializer=TekHighspeedServer__pb2.NormalizedReply.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "Tekscope.NormalizedData", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class NormalizedData(object):
    """Normalized Waveform (this is fully defined)"""

    @staticmethod
    def GetWaveform(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_stream(
            request,
            target,
            "/Tekscope.NormalizedData/GetWaveform",
            TekHighspeedServer__pb2.WaveformRequest.SerializeToString,
            TekHighspeedServer__pb2.NormalizedReply.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetHeader(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/Tekscope.NormalizedData/GetHeader",
            TekHighspeedServer__pb2.WaveformRequest.SerializeToString,
            TekHighspeedServer__pb2.NormalizedReply.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )


class NativeDataStub(object):
    """Returns internal data (which is smaller)
    However it comes in a byte stream that must be interpreted
    by the client based on the header definition.
    """

    def __init__(self, channel: grpc.Channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetWaveform = channel.unary_stream(
            "/Tekscope.NativeData/GetWaveform",
            request_serializer=TekHighspeedServer__pb2.WaveformRequest.SerializeToString,
            response_deserializer=TekHighspeedServer__pb2.RawReply.FromString,
        )
        self.GetHeader = channel.unary_unary(
            "/Tekscope.NativeData/GetHeader",
            request_serializer=TekHighspeedServer__pb2.WaveformRequest.SerializeToString,
            response_deserializer=TekHighspeedServer__pb2.RawReply.FromString,
        )


class NativeDataServicer(object):
    """Returns internal data (which is smaller)
    However it comes in a byte stream that must be interpreted
    by the client based on the header definition.
    """

    def GetWaveform(self, request, context):
        """Query Waveform Samples."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetHeader(self, request, context):
        """Query Header information (needed to interpret samples)"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_NativeDataServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "GetWaveform": grpc.unary_stream_rpc_method_handler(
            servicer.GetWaveform,
            request_deserializer=TekHighspeedServer__pb2.WaveformRequest.FromString,
            response_serializer=TekHighspeedServer__pb2.RawReply.SerializeToString,
        ),
        "GetHeader": grpc.unary_unary_rpc_method_handler(
            servicer.GetHeader,
            request_deserializer=TekHighspeedServer__pb2.WaveformRequest.FromString,
            response_serializer=TekHighspeedServer__pb2.RawReply.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "Tekscope.NativeData", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class NativeData(object):
    """Returns internal data (which is smaller)
    However it comes in a byte stream that must be interpreted
    by the client based on the header definition.
    """

    @staticmethod
    def GetWaveform(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_stream(
            request,
            target,
            "/Tekscope.NativeData/GetWaveform",
            TekHighspeedServer__pb2.WaveformRequest.SerializeToString,
            TekHighspeedServer__pb2.RawReply.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetHeader(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/Tekscope.NativeData/GetHeader",
            TekHighspeedServer__pb2.WaveformRequest.SerializeToString,
            TekHighspeedServer__pb2.RawReply.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



"""Tektronix High Speed Interface.

Provides access to commonly imported items from the `TekHSI` package.
"""

from importlib.metadata import version

from tekhsi._tek_highspeed_server_pb2 import WaveformHeader  # pylint: disable= no-name-in-module
from tekhsi.helpers import PACKAGE_NAME
from tekhsi.tek_hsi_connect import AcqWaitOn, TekHSIConnect

# Read version from installed package.
__version__ = version(PACKAGE_NAME)

__all__ = [
    "TekHSIConnect",
    "AcqWaitOn",
    "WaveformHeader",
]



from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

CONNECTSTATUS_INUSE_FAILURE: ConnectStatus
CONNECTSTATUS_NOT_CONNECTED_FAILURE: ConnectStatus
CONNECTSTATUS_OUTSIDE_SEQUENCE_FAILURE: ConnectStatus
CONNECTSTATUS_SUCCESS: ConnectStatus
CONNECTSTATUS_TIMEOUT_FAILURE: ConnectStatus
CONNECTSTATUS_UNKNOWN_FAILURE: ConnectStatus
CONNECTSTATUS_UNSPECIFIED: ConnectStatus
DESCRIPTOR: _descriptor.FileDescriptor
WFMPAIRTYPE_NONE: WfmPairType
WFMPAIRTYPE_PAIR: WfmPairType
WFMPAIRTYPE_UNSPECIFIED: WfmPairType
WFMREPLYSTATUS_NO_CONNECTION_FAILURE: WfmReplyStatus
WFMREPLYSTATUS_OUTSIDE_SEQUENCE_FAILURE: WfmReplyStatus
WFMREPLYSTATUS_SOURCENAME_MISSING_FAILURE: WfmReplyStatus
WFMREPLYSTATUS_SUCCESS: WfmReplyStatus
WFMREPLYSTATUS_TYPE_MISMATCH_FAILURE: WfmReplyStatus
WFMREPLYSTATUS_UNSPECIFIED: WfmReplyStatus
WFMTYPE_ANALOG_16: WfmType
WFMTYPE_ANALOG_16_IQ: WfmType
WFMTYPE_ANALOG_32_IQ: WfmType
WFMTYPE_ANALOG_8: WfmType
WFMTYPE_ANALOG_FLOAT: WfmType
WFMTYPE_DIGITAL_16: WfmType
WFMTYPE_DIGITAL_8: WfmType
WFMTYPE_UNSPECIFIED: WfmType

class AvailableNamesReply(_message.Message):
    __slots__ = ["status", "symbolnames"]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    SYMBOLNAMES_FIELD_NUMBER: _ClassVar[int]
    status: ConnectStatus
    symbolnames: _containers.RepeatedScalarFieldContainer[str]
    def __init__(
        self,
        status: _Optional[_Union[ConnectStatus, str]] = ...,
        symbolnames: _Optional[_Iterable[str]] = ...,
    ) -> None: ...

class ConnectReply(_message.Message):
    __slots__ = ["status"]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    status: ConnectStatus
    def __init__(self, status: _Optional[_Union[ConnectStatus, str]] = ...) -> None: ...

class ConnectRequest(_message.Message):
    __slots__ = ["name"]
    NAME_FIELD_NUMBER: _ClassVar[int]
    name: str
    def __init__(self, name: _Optional[str] = ...) -> None: ...

class NormalizedReply(_message.Message):
    __slots__ = ["headerordata", "status"]
    class DataOrHeaderAccess(_message.Message):
        __slots__ = ["chunk", "header"]
        CHUNK_FIELD_NUMBER: _ClassVar[int]
        HEADER_FIELD_NUMBER: _ClassVar[int]
        chunk: NormalizedReply.WaveformSampleChunk
        header: WaveformHeader
        def __init__(
            self,
            header: _Optional[_Union[WaveformHeader, _Mapping]] = ...,
            chunk: _Optional[_Union[NormalizedReply.WaveformSampleChunk, _Mapping]] = ...,
        ) -> None: ...

    class WaveformSampleChunk(_message.Message):
        __slots__ = ["data"]
        DATA_FIELD_NUMBER: _ClassVar[int]
        data: _containers.RepeatedScalarFieldContainer[float]
        def __init__(self, data: _Optional[_Iterable[float]] = ...) -> None: ...

    HEADERORDATA_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    headerordata: NormalizedReply.DataOrHeaderAccess
    status: WfmReplyStatus
    def __init__(
        self,
        status: _Optional[_Union[WfmReplyStatus, str]] = ...,
        headerordata: _Optional[_Union[NormalizedReply.DataOrHeaderAccess, _Mapping]] = ...,
    ) -> None: ...

class RawReply(_message.Message):
    __slots__ = ["headerordata", "status"]
    class DataOrHeaderAccess(_message.Message):
        __slots__ = ["chunk", "header"]
        CHUNK_FIELD_NUMBER: _ClassVar[int]
        HEADER_FIELD_NUMBER: _ClassVar[int]
        chunk: RawReply.WaveformSampleByteChunk
        header: WaveformHeader
        def __init__(
            self,
            header: _Optional[_Union[WaveformHeader, _Mapping]] = ...,
            chunk: _Optional[_Union[RawReply.WaveformSampleByteChunk, _Mapping]] = ...,
        ) -> None: ...

    class WaveformSampleByteChunk(_message.Message):
        __slots__ = ["data"]
        DATA_FIELD_NUMBER: _ClassVar[int]
        data: bytes
        def __init__(self, data: _Optional[bytes] = ...) -> None: ...

    HEADERORDATA_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    headerordata: RawReply.DataOrHeaderAccess
    status: WfmReplyStatus
    def __init__(
        self,
        status: _Optional[_Union[WfmReplyStatus, str]] = ...,
        headerordata: _Optional[_Union[RawReply.DataOrHeaderAccess, _Mapping]] = ...,
    ) -> None: ...

class WaveformHeader(_message.Message):
    __slots__ = [
        "bitmask",
        "chunksize",
        "dataid",
        "hasdata",
        "horizontalUnits",
        "horizontalfractionalzeroindex",
        "horizontalspacing",
        "horizontalzeroindex",
        "iq_centerFrequency",
        "iq_fftLength",
        "iq_rbw",
        "iq_span",
        "iq_windowType",
        "noofsamples",
        "pairtype",
        "sourcename",
        "sourcewidth",
        "transid",
        "verticaloffset",
        "verticalspacing",
        "verticalunits",
        "wfmtype",
    ]
    BITMASK_FIELD_NUMBER: _ClassVar[int]
    CHUNKSIZE_FIELD_NUMBER: _ClassVar[int]
    DATAID_FIELD_NUMBER: _ClassVar[int]
    HASDATA_FIELD_NUMBER: _ClassVar[int]
    HORIZONTALFRACTIONALZEROINDEX_FIELD_NUMBER: _ClassVar[int]
    HORIZONTALSPACING_FIELD_NUMBER: _ClassVar[int]
    HORIZONTALUNITS_FIELD_NUMBER: _ClassVar[int]
    HORIZONTALZEROINDEX_FIELD_NUMBER: _ClassVar[int]
    IQ_CENTERFREQUENCY_FIELD_NUMBER: _ClassVar[int]
    IQ_FFTLENGTH_FIELD_NUMBER: _ClassVar[int]
    IQ_RBW_FIELD_NUMBER: _ClassVar[int]
    IQ_SPAN_FIELD_NUMBER: _ClassVar[int]
    IQ_WINDOWTYPE_FIELD_NUMBER: _ClassVar[int]
    NOOFSAMPLES_FIELD_NUMBER: _ClassVar[int]
    PAIRTYPE_FIELD_NUMBER: _ClassVar[int]
    SOURCENAME_FIELD_NUMBER: _ClassVar[int]
    SOURCEWIDTH_FIELD_NUMBER: _ClassVar[int]
    TRANSID_FIELD_NUMBER: _ClassVar[int]
    VERTICALOFFSET_FIELD_NUMBER: _ClassVar[int]
    VERTICALSPACING_FIELD_NUMBER: _ClassVar[int]
    VERTICALUNITS_FIELD_NUMBER: _ClassVar[int]
    WFMTYPE_FIELD_NUMBER: _ClassVar[int]
    bitmask: int
    chunksize: int
    dataid: int
    hasdata: bool
    horizontalUnits: str
    horizontalfractionalzeroindex: float
    horizontalspacing: float
    horizontalzeroindex: float
    iq_centerFrequency: float
    iq_fftLength: float
    iq_rbw: float
    iq_span: float
    iq_windowType: str
    noofsamples: int
    pairtype: WfmPairType
    sourcename: str
    sourcewidth: int
    transid: int
    verticaloffset: float
    verticalspacing: float
    verticalunits: str
    wfmtype: WfmType
    def __init__(
        self,
        sourcename: _Optional[str] = ...,
        sourcewidth: _Optional[int] = ...,
        dataid: _Optional[int] = ...,
        transid: _Optional[int] = ...,
        horizontalUnits: _Optional[str] = ...,
        horizontalspacing: _Optional[float] = ...,
        horizontalzeroindex: _Optional[float] = ...,
        horizontalfractionalzeroindex: _Optional[float] = ...,
        noofsamples: _Optional[int] = ...,
        chunksize: _Optional[int] = ...,
        wfmtype: _Optional[_Union[WfmType, str]] = ...,
        bitmask: _Optional[int] = ...,
        pairtype: _Optional[_Union[WfmPairType, str]] = ...,
        verticalunits: _Optional[str] = ...,
        verticalspacing: _Optional[float] = ...,
        verticaloffset: _Optional[float] = ...,
        iq_centerFrequency: _Optional[float] = ...,
        iq_fftLength: _Optional[float] = ...,
        iq_rbw: _Optional[float] = ...,
        iq_span: _Optional[float] = ...,
        iq_windowType: _Optional[str] = ...,
        hasdata: bool = ...,
    ) -> None: ...

class WaveformRequest(_message.Message):
    __slots__ = ["chunksize", "sourcename"]
    CHUNKSIZE_FIELD_NUMBER: _ClassVar[int]
    SOURCENAME_FIELD_NUMBER: _ClassVar[int]
    chunksize: int
    sourcename: str
    def __init__(
        self, sourcename: _Optional[str] = ..., chunksize: _Optional[int] = ...
    ) -> None: ...

class ConnectStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class WfmReplyStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class WfmPairType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []

class WfmType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: TekHighspeedServer.proto
"""Generated protocol buffer code."""

from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x18TekHighspeedServer.proto\x12\x08Tekscope"\x1e\n\x0e\x43onnectRequest\x12\x0c\n\x04name\x18\x01 \x01(\t"7\n\x0c\x43onnectReply\x12\'\n\x06status\x18\x01 \x01(\x0e\x32\x17.Tekscope.ConnectStatus"S\n\x13\x41vailableNamesReply\x12\'\n\x06status\x18\x01 \x01(\x0e\x32\x17.Tekscope.ConnectStatus\x12\x13\n\x0bsymbolnames\x18\x02 \x03(\t"8\n\x0fWaveformRequest\x12\x12\n\nsourcename\x18\x01 \x01(\t\x12\x11\n\tchunksize\x18\x02 \x01(\r"\x9b\x04\n\x0eWaveformHeader\x12\x12\n\nsourcename\x18\x01 \x01(\t\x12\x13\n\x0bsourcewidth\x18\x02 \x01(\r\x12\x0e\n\x06\x64\x61taid\x18\x03 \x01(\x04\x12\x0f\n\x07transid\x18\x04 \x01(\x04\x12\x17\n\x0fhorizontalUnits\x18\x05 \x01(\t\x12\x19\n\x11horizontalspacing\x18\x06 \x01(\x01\x12\x1b\n\x13horizontalzeroindex\x18\x07 \x01(\x01\x12%\n\x1dhorizontalfractionalzeroindex\x18\x08 \x01(\x01\x12\x13\n\x0bnoofsamples\x18\t \x01(\x04\x12\x11\n\tchunksize\x18\n \x01(\r\x12"\n\x07wfmtype\x18\x0b \x01(\x0e\x32\x11.Tekscope.WfmType\x12\x0f\n\x07\x62itmask\x18\x0c \x01(\r\x12\'\n\x08pairtype\x18\r \x01(\x0e\x32\x15.Tekscope.WfmPairType\x12\x15\n\rverticalunits\x18\x0e \x01(\t\x12\x17\n\x0fverticalspacing\x18\x0f \x01(\x01\x12\x16\n\x0everticaloffset\x18\x10 \x01(\x01\x12\x1a\n\x12iq_centerFrequency\x18\x11 \x01(\x01\x12\x14\n\x0ciq_fftLength\x18\x12 \x01(\x01\x12\x0e\n\x06iq_rbw\x18\x13 \x01(\x01\x12\x0f\n\x07iq_span\x18\x14 \x01(\x01\x12\x15\n\riq_windowType\x18\x15 \x01(\t\x12\x0f\n\x07hasdata\x18\x16 \x01(\x08"\xb4\x02\n\x0fNormalizedReply\x12(\n\x06status\x18\x01 \x01(\x0e\x32\x18.Tekscope.WfmReplyStatus\x12\x42\n\x0cheaderordata\x18\x02 \x01(\x0b\x32,.Tekscope.NormalizedReply.DataOrHeaderAccess\x1a\'\n\x13WaveformSampleChunk\x12\x10\n\x04\x64\x61ta\x18\x01 \x03(\x02\x42\x02\x10\x01\x1a\x89\x01\n\x12\x44\x61taOrHeaderAccess\x12*\n\x06header\x18\x01 \x01(\x0b\x32\x18.Tekscope.WaveformHeaderH\x00\x12>\n\x05\x63hunk\x18\x02 \x01(\x0b\x32-.Tekscope.NormalizedReply.WaveformSampleChunkH\x00\x42\x07\n\x05value"\xa3\x02\n\x08RawReply\x12(\n\x06status\x18\x01 \x01(\x0e\x32\x18.Tekscope.WfmReplyStatus\x12;\n\x0cheaderordata\x18\x02 \x01(\x0b\x32%.Tekscope.RawReply.DataOrHeaderAccess\x1a\'\n\x17WaveformSampleByteChunk\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\x1a\x86\x01\n\x12\x44\x61taOrHeaderAccess\x12*\n\x06header\x18\x01 \x01(\x0b\x32\x18.Tekscope.WaveformHeaderH\x00\x12;\n\x05\x63hunk\x18\x02 \x01(\x0b\x32*.Tekscope.RawReply.WaveformSampleByteChunkH\x00\x42\x07\n\x05value*\x85\x02\n\rConnectStatus\x12\x1d\n\x19\x43ONNECTSTATUS_UNSPECIFIED\x10\x00\x12\x19\n\x15\x43ONNECTSTATUS_SUCCESS\x10\x01\x12\'\n#CONNECTSTATUS_NOT_CONNECTED_FAILURE\x10\x02\x12*\n&CONNECTSTATUS_OUTSIDE_SEQUENCE_FAILURE\x10\x03\x12!\n\x1d\x43ONNECTSTATUS_TIMEOUT_FAILURE\x10\x04\x12\x1f\n\x1b\x43ONNECTSTATUS_INUSE_FAILURE\x10\x05\x12!\n\x1d\x43ONNECTSTATUS_UNKNOWN_FAILURE\x10\x06*\xfc\x01\n\x0eWfmReplyStatus\x12\x1e\n\x1aWFMREPLYSTATUS_UNSPECIFIED\x10\x00\x12\x1a\n\x16WFMREPLYSTATUS_SUCCESS\x10\x01\x12-\n)WFMREPLYSTATUS_SOURCENAME_MISSING_FAILURE\x10\x02\x12+\n\'WFMREPLYSTATUS_OUTSIDE_SEQUENCE_FAILURE\x10\x03\x12(\n$WFMREPLYSTATUS_NO_CONNECTION_FAILURE\x10\x04\x12(\n$WFMREPLYSTATUS_TYPE_MISMATCH_FAILURE\x10\x05*V\n\x0bWfmPairType\x12\x1b\n\x17WFMPAIRTYPE_UNSPECIFIED\x10\x00\x12\x14\n\x10WFMPAIRTYPE_NONE\x10\x01\x12\x14\n\x10WFMPAIRTYPE_PAIR\x10\x02*\xcc\x01\n\x07WfmType\x12\x17\n\x13WFMTYPE_UNSPECIFIED\x10\x00\x12\x14\n\x10WFMTYPE_ANALOG_8\x10\x01\x12\x15\n\x11WFMTYPE_ANALOG_16\x10\x02\x12\x18\n\x14WFMTYPE_ANALOG_FLOAT\x10\x03\x12\x15\n\x11WFMTYPE_DIGITAL_8\x10\x04\x12\x16\n\x12WFMTYPE_DIGITAL_16\x10\x05\x12\x18\n\x14WFMTYPE_ANALOG_16_IQ\x10\x06\x12\x18\n\x14WFMTYPE_ANALOG_32_IQ\x10\x07\x32\xbf\x03\n\x07\x43onnect\x12=\n\x07\x43onnect\x12\x18.Tekscope.ConnectRequest\x1a\x16.Tekscope.ConnectReply"\x00\x12@\n\nDisconnect\x12\x18.Tekscope.ConnectRequest\x1a\x16.Tekscope.ConnectReply"\x00\x12H\n\x12RequestNewSequence\x12\x18.Tekscope.ConnectRequest\x1a\x16.Tekscope.ConnectReply"\x00\x12R\n\x15RequestAvailableNames\x12\x18.Tekscope.ConnectRequest\x1a\x1d.Tekscope.AvailableNamesReply"\x00\x12G\n\x11WaitForDataAccess\x12\x18.Tekscope.ConnectRequest\x1a\x16.Tekscope.ConnectReply"\x00\x12L\n\x16\x46inishedWithDataAccess\x12\x18.Tekscope.ConnectRequest\x1a\x16.Tekscope.ConnectReply"\x00\x32\x9e\x01\n\x0eNormalizedData\x12G\n\x0bGetWaveform\x12\x19.Tekscope.WaveformRequest\x1a\x19.Tekscope.NormalizedReply"\x00\x30\x01\x12\x43\n\tGetHeader\x12\x19.Tekscope.WaveformRequest\x1a\x19.Tekscope.NormalizedReply"\x00\x32\x8c\x01\n\nNativeData\x12@\n\x0bGetWaveform\x12\x19.Tekscope.WaveformRequest\x1a\x12.Tekscope.RawReply"\x00\x30\x01\x12<\n\tGetHeader\x12\x19.Tekscope.WaveformRequest\x1a\x12.Tekscope.RawReply"\x00\x62\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "TekHighspeedServer_pb2", globals())
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _NORMALIZEDREPLY_WAVEFORMSAMPLECHUNK.fields_by_name["data"]._options = None
    _NORMALIZEDREPLY_WAVEFORMSAMPLECHUNK.fields_by_name["data"]._serialized_options = b"\020\001"
    _CONNECTSTATUS._serialized_start = 1418
    _CONNECTSTATUS._serialized_end = 1679
    _WFMREPLYSTATUS._serialized_start = 1682
    _WFMREPLYSTATUS._serialized_end = 1934
    _WFMPAIRTYPE._serialized_start = 1936
    _WFMPAIRTYPE._serialized_end = 2022
    _WFMTYPE._serialized_start = 2025
    _WFMTYPE._serialized_end = 2229
    _CONNECTREQUEST._serialized_start = 38
    _CONNECTREQUEST._serialized_end = 68
    _CONNECTREPLY._serialized_start = 70
    _CONNECTREPLY._serialized_end = 125
    _AVAILABLENAMESREPLY._serialized_start = 127
    _AVAILABLENAMESREPLY._serialized_end = 210
    _WAVEFORMREQUEST._serialized_start = 212
    _WAVEFORMREQUEST._serialized_end = 268
    _WAVEFORMHEADER._serialized_start = 271
    _WAVEFORMHEADER._serialized_end = 810
    _NORMALIZEDREPLY._serialized_start = 813
    _NORMALIZEDREPLY._serialized_end = 1121
    _NORMALIZEDREPLY_WAVEFORMSAMPLECHUNK._serialized_start = 942
    _NORMALIZEDREPLY_WAVEFORMSAMPLECHUNK._serialized_end = 981
    _NORMALIZEDREPLY_DATAORHEADERACCESS._serialized_start = 984
    _NORMALIZEDREPLY_DATAORHEADERACCESS._serialized_end = 1121
    _RAWREPLY._serialized_start = 1124
    _RAWREPLY._serialized_end = 1415
    _RAWREPLY_WAVEFORMSAMPLEBYTECHUNK._serialized_start = 1239
    _RAWREPLY_WAVEFORMSAMPLEBYTECHUNK._serialized_end = 1278
    _RAWREPLY_DATAORHEADERACCESS._serialized_start = 1281
    _RAWREPLY_DATAORHEADERACCESS._serialized_end = 1415
    _CONNECT._serialized_start = 2232
    _CONNECT._serialized_end = 2679
    _NORMALIZEDDATA._serialized_start = 2682
    _NORMALIZEDDATA._serialized_end = 2840
    _NATIVEDATA._serialized_start = 2843
    _NATIVEDATA._serialized_end = 2983
# @@protoc_insertion_point(module_scope)



"""Module for connecting to Tektronix instruments and retrieving waveform data using gRPC."""

import contextlib
import threading
import time
import uuid

from atexit import register
from enum import Enum
from typing import ClassVar, Dict, List, Optional

import grpc
import numpy as np

from tm_data_types import (
    AnalogWaveform,
    DigitalWaveform,
    IQWaveform,
    IQWaveformMetaInfo,
    Waveform,
)

from tekhsi._tek_highspeed_server_pb2 import (  # pylint: disable=no-name-in-module
    ConnectRequest,
    WaveformHeader,
    WaveformRequest,
)
from tekhsi._tek_highspeed_server_pb2_grpc import ConnectStub, NativeDataStub
from tekhsi.helpers.functions import print_with_timestamp


class AcqWaitOn(Enum):
    """This defines enumeration for wait_for_data."""

    NextAcq = 1
    """Wait for the next acquisition."""
    Time = 2
    """Wait for a specific time."""
    AnyAcq = 3
    """Wait for any acquisition."""
    NewData = 4
    """Wait for new data."""


class TekHSIConnect:  # pylint:disable=too-many-instance-attributes
    """Support for Tektronix highspeed data API.

    - this API is intended to aid in retrieving data from instruments as fast as possible.
    """

    _connections: ClassVar[Dict[str, "TekHSIConnect"]] = {}

    ################################################################################################
    # Magic Methods
    ################################################################################################
    def __init__(self, url: str, activesymbols=None, callback=None, data_filter=None):
        """Initialize a connection to a Tektronix instrument using gRPC.

        Args:
           url (str): The URL of the gRPC server.
           activesymbols (list, optional): List of active symbols to monitor.
           callback (function, optional): Callback function to handle incoming data.
           data_filter (function, optional): Filter function to apply to incoming data.
        """
        self.previous_headers = []
        self.chunksize = 80000
        self.url = url
        self.v_datatypes = {1: np.int8, 2: np.int16, 4: np.float32, 8: np.double}
        self.iq_datatypes = {1: np.int8, 2: np.int16, 4: np.int32}
        self.d_datatypes = {1: np.int8}
        self.channel = grpc.insecure_channel(url)
        self.clientname = str(uuid.uuid4())
        self.connection = ConnectStub(self.channel)
        self.native = NativeDataStub(self.channel)
        self.thread_active = True
        self._callback = callback
        self._holding_scope_open = False
        self._verbose = False
        self._instrument = False
        self._cachedataenabled = True
        self._lock = threading.Lock()
        self._lock_getdata = threading.Lock()
        self._lock_filter = threading.Lock()
        self._datacache = {}
        self._headers = {}
        self._connect()
        self._connected = True
        self._recordlength = 0
        self._acqcount = 0
        self._acqtime = -1
        self._filter = data_filter
        self._lastacqseen = self._acqcount
        self._wait_for_data_count = 0
        self._start_time = time.time()
        self._wait_for_data_holds_lock = False
        self._in_wait_for_data = False
        self._sum_transfer_time = 0
        self._sum_acq_time = 0
        self._sum_data_rate = 0
        self._sum_count = 0
        self._is_exiting = False
        self._prev_data_id = -1

        TekHSIConnect._connections[self.clientname] = self

        if not activesymbols:
            self.activesymbols = self._available_symbols()
        else:
            self.activesymbols = [x.lower() for x in activesymbols]

        self.thread = threading.Thread(target=self._run, args=())
        self.thread.daemon = True
        self.thread.start()

    def __enter__(self):
        """Enter the runtime context related to this object.

        Returns:
            The object itself.
        """
        # Required for "with" command to work with this class
        if self.verbose:
            print_with_timestamp("enter()")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the runtime context related to this object.

        Args:
            exc_type: The exception type.
            exc_val: The exception value.
            exc_tb: The traceback object.
        """
        # Required for "with" command to work with this class

        self._is_exiting = True

        if self.verbose:
            print_with_timestamp("exit()")

        self.close()

        if self._instrument and self._sum_count > 0:
            print_with_timestamp(
                f"Average Update Rate:{(1 / (self._sum_acq_time / self._sum_count)):.2f}, "
                f"Data Rate:{(self._sum_data_rate / self._sum_count):.2f}Mbs",
            )

    ################################################################################################
    # Properties - Private and Public
    ################################################################################################
    @property
    def available_symbols(self) -> List[str]:
        """Returns the list of available symbols on the instrument.

        Returns:
            List[str]: List of available symbols.
        """
        return self._available_symbols()

    @property
    def current_time(self) -> float:
        """This property returns time relative to the connection to the gRPC client.

        Returns:
            float: current time relative to the start time of the gRPC client
        """
        return time.time() - self._start_time

    @property
    def instrumentation_enabled(self) -> bool:
        """Indicates if instrumentation is enabled.

        Returns:
            True if instrumentation is enabled, False otherwise.
        """
        return self._instrument

    @instrumentation_enabled.setter
    def instrumentation_enabled(self, value: bool):
        """Sets the instrumentation enabled state.

        Args:
            value: True to enable instrumentation, False to disable.
        """
        self._instrument = value

    @property
    def source_names(self) -> List[str]:
        """Returns the list of names of sources on the instrument.

        Returns:
            List[str]: list of sources
        """
        return self.activesymbols

    @property
    def verbose(self) -> bool:
        """Indicates if verbose mode is enabled.

        Returns:
            True if verbose mode is enabled, False otherwise.
        """
        return self._verbose

    @verbose.setter
    def verbose(self, value: bool):
        """Sets the verbose mode state.

        Args:
            value: True to enable verbose mode, False to disable.
        """
        self._verbose = value

    ################################################################################################
    # Context Manager Methods
    ################################################################################################
    @contextlib.contextmanager
    def access_data(self, on: AcqWaitOn = AcqWaitOn.NewData, after: float = -1):
        """Grants access to data.

        Must be called as part of a with command to grant access for get_data().

        Args:
            on: Criterion for acceptance. Defaults to AcqWaitOn.NewData.
            after: Additional criterion when AcqWaitOn.Time is passed. Defaults to -1.
        """
        try:
            self.wait_for_data(on, after)
            yield self
        finally:
            self.done_with_data()

    ################################################################################################
    # Public Methods
    ################################################################################################
    @staticmethod
    def any_acq(
        previous_header: Dict[str, WaveformHeader],  # noqa: ARG004
        current_header: Dict[str, WaveformHeader],  # noqa: ARG004
    ) -> bool:
        """Prebuilt acq acceptance filter that accepts all new acqs.

        Args:
            previous_header: Previous header.
            current_header: Current header.

        Returns:
            True if the acquisition is accepted, False otherwise.
        """
        return True

    @staticmethod
    def any_horizontal_change(
        previous_header: Dict[str, WaveformHeader],
        current_header: Dict[str, WaveformHeader],
    ) -> bool:
        """Acq acceptance filter that accepts only acqs with changes to horizontal settings.

        Args:
            previous_header (dict[str, WaveformHeader]): Previous header dictionary.
            current_header (dict[str, WaveformHeader]): Current header dictionary.

        Returns:
            True if the acquisition is accepted, False otherwise.
        """
        for key, cur in current_header.items():
            if key not in previous_header:
                return True
            prev = previous_header[key]
            if prev is None and cur is not None:
                return True
            if prev is not None and (
                prev.noofsamples != cur.noofsamples
                or prev.horizontalspacing != cur.horizontalspacing
                or prev.horizontalzeroindex != cur.horizontalzeroindex
            ):
                return True
        return False

    @staticmethod
    def any_vertical_change(
        previous_header: Dict[str, WaveformHeader],
        current_header: Dict[str, WaveformHeader],
    ) -> bool:
        """Prebuilt acq acceptance filter that accepts only acqs with changes to vertical settings.

        Args:
            previous_header (dict[str, WaveformHeader]): Previous header dictionary.
            current_header (dict[str, WaveformHeader]): Current header dictionary.

        Returns:
            True if the acquisition is accepted, False otherwise.
        """
        for key, cur in current_header.items():
            if key not in previous_header:
                return True
            prev = previous_header[key]
            if prev is not None and (
                prev.verticalspacing != cur.verticalspacing
                or prev.verticaloffset != cur.verticaloffset
            ):
                return True
            if prev is None and cur is not None:
                return True
        return False

    def active_symbols(self, symbols: List[str]) -> None:
        """Sets symbols to consider moving from instrument into data cache.

        Args:
            symbols (List[str]): list of symbols to be moved
        """
        self.activesymbols = symbols

    def close(self):
        """Close and clean up gRPC connection."""
        if not self._connected:
            return

        if self.verbose:
            print_with_timestamp("close")

        # This will force the scope to give the background
        # thread access to data. That should cause it to exit.
        try:
            self.thread_active = False
            self.force_sequence()
            # Wait for thread to exit
            self.thread.join(20.0)
        except RuntimeError as error:
            # Handle specific exceptions related to threading
            if self.verbose:
                msg = f"Thread error: {error}"
                print_with_timestamp(msg)

        try:
            # Take this connection out of the connection list
            if self.clientname in TekHSIConnect._available_symbols(self):
                del TekHSIConnect._available_symbols[self.clientname]  # pylint:disable=unsupported-delete-operation
        except KeyError as error:
            # Handle specific exception if the key is not found
            if self.verbose:
                msg = f"Key error: {error}"
                print_with_timestamp(msg)

        if self.verbose:
            print_with_timestamp("disconnect")

        # disconnect from the instrument
        self._disconnect()

    @staticmethod
    def data_arrival(waveforms: List[Waveform]) -> None:  # noqa: ARG004
        """Available to be overridden if user wants to create a derived class.

        This method will be called on every accepted acq.

        Args:
            waveforms: List of waveforms.
        """
        return

    def done_with_data(self):
        """Releases the acquisition after accessing the required data."""
        if not self._cachedataenabled:
            return

        if self._wait_for_data_count <= 0:
            if self.verbose:
                print_with_timestamp("** done_with_data called when no wait_for_data pending")
            return

        self._wait_for_data_count -= 1
        self._lastacqseen = self._acqcount
        self._done_with_data_release_lock()
        return

    def force_sequence(self) -> None:
        """force_sequence asks the instrument to please give us access.

        to the current acquisition data. This is useful when connecting to a stopped instrument to
        get access to the currently available data. Otherwise, the API will wait until the next
        acquisition.
        """
        if self.verbose:
            print_with_timestamp("force sequence")
        request = ConnectRequest(name=self.clientname)
        self.connection.RequestNewSequence(request)

    def get_data(self, name: str) -> Optional[Waveform]:
        """Returns the saved data of the previous acquisition with the data item.

        of the requested name.

        Args:
            name: Name of the data item.

        Returns:
            The waveform data or None if caching is off or data is not found.
        """
        if not self._cachedataenabled:
            return None  # Return None if caching off.

        self._lock_getdata.acquire()
        try:
            retval = self._datacache.get(
                name.lower(),
                None,
            )  # Return None if cached data is not found
        finally:
            self._lock_getdata.release()
        return retval

    def set_acq_filter(self, acq_filter):
        """Sets rules for acquisitions that are accepted and forwarded.

        This is to allow only import data changes to be passed to the callback or saved to backing
        store.

        Args:
            acq_filter  (function): A function that takes two headers and returns True if the data
        """
        if acq_filter is None:
            raise ValueError("Filter cannot be None")

        self._lock_filter.acquire()
        self._filter = acq_filter
        self._lastacqseen = self._acqcount
        self._lock_filter.release()

    def wait_for_data(self, on: AcqWaitOn = AcqWaitOn.NewData, after: float = -1):
        """Waits until specified acquisition criterion is met.

        Args:
            on (AcqWaitOn, optional): criterion for acceptance. Defaults to AcqWaitOn.NewData.
            after (float, optional): additional criterion when AcqWaitOn.Time is passed.
        """
        if not self._cachedataenabled:
            return

        if on == AcqWaitOn.AnyAcq:
            self._wait_for_any_acq()
        elif on == AcqWaitOn.NextAcq:
            self._wait_for_next_acq()
        elif on == AcqWaitOn.Time:
            self._wait_for_acq_time(after)
        elif on == AcqWaitOn.NewData:
            self._wait_for_new_data()

    ################################################################################################
    # Private Methods
    ################################################################################################

    @staticmethod
    def _acq_id(headers: List[WaveformHeader]):
        """Retrieve the data ID from the first header in the list.

        Args:
            headers (List[WaveformHeader]): List of waveform headers.

        Returns:
            The data ID of the first header, or None if the list is empty.
        """
        for header in headers:
            return header.dataid
        return None

    def _available_symbols(self) -> List[str]:
        """Returns the list of available channels.

        Returns:
            List of available channels.
        """
        request = ConnectRequest(name=self.clientname)
        response = self.connection.RequestAvailableNames(request)
        return response.symbolnames

    def _connect(self) -> None:
        """Request connect to the gRCP server."""
        if self.verbose:
            print_with_timestamp("connect")
        request = ConnectRequest(name=self.clientname)
        self.connection.Connect(request)

    def _disconnect(self) -> None:
        """Disconnect from gRPC server."""
        if not self._connected:
            return
        self._connected = False
        if self.verbose:
            print_with_timestamp("disconnect")
        request = ConnectRequest(name=self.clientname)
        self.connection.Disconnect(request)

    def _done_with_data_release_lock(self):
        """Releases the lock after accessing the required data."""
        if self._wait_for_data_holds_lock:
            self._lock.release()
            self._wait_for_data_holds_lock = False

    def _instrumentation(self, acqtime, transfertime, datasize, datawidth) -> None:
        """Prints the performance information for debugging.

        Args:
            acqtime: Acquisition time.
            transfertime: Transfer time.
            datasize: Data size.
            datawidth: Data width.
        """
        if self._instrument and self._connected and not self._is_exiting:
            self._sum_acq_time += acqtime
            self._sum_transfer_time += transfertime
            self._sum_data_rate += (datasize * 8 / 1e6) / transfertime
            self._sum_count += 1
            print(
                f"UpdateRate:{(1 / acqtime):.2f},"
                f"Data Rate:{((datasize * 8 / 1e6) / transfertime):.2f}Mbs,Data Width:{datawidth}",
            )

    @staticmethod
    def _is_header_value(header: WaveformHeader) -> bool:
        """Check if the header has valid data.

        Args:
            header (WaveformHeader): The waveform header to check.

        Returns:
            bool: True if the header has valid data, False otherwise.
        """
        return (
            header is not None
            and header.noofsamples > 0
            and header.sourcewidth in {1, 2, 4}
            and header.hasdata
        )

    def _finished_with_data_access(self) -> None:
        """Releases access to instrument data - this is required
        to allow the instrument to continue acquiring
        """
        if not self._in_wait_for_data:
            return

        if self.verbose:
            print_with_timestamp("finished_with_data_access")

        request = ConnectRequest(name=self.clientname)
        self.connection.FinishedWithDataAccess(request)

    def _read_header(self, name: str) -> WaveformHeader:
        """Reads header for the named source.

        Args:
            name (str): name of header

        Returns:
            WaveformHeader: the description of the properties of the specified waveform
        """
        if self.verbose:
            print(f"{name}:read header")
        request = WaveformRequest(sourcename=name, chunksize=self.chunksize)
        response = self.native.GetHeader(request)
        return response.headerordata.header

    def _read_headers(self, headers, header_dict: dict):
        """Reads headers for the active symbols.

        Args:
            headers (list): list of headers
            header_dict (dict): dictionary of headers
        """
        symbols = self.activesymbols
        n = len(symbols)
        for index in range(n):
            symbol = symbols[index]
            header = self._read_header(symbol)
            if self._is_header_value(header):
                headers.append(header)
                header_dict[header.sourcename] = header

    # pylint: disable= too-many-locals
    def _read_waveform(self, header: WaveformHeader):  # noqa: C901,PLR0912,PLR0915
        """Reads the analog waveform associated with the passed header.

        Args:
            header (WaveformHeader): the header of the source to read data from

        Returns:
            Waveform: contains definition and data for the specified source
        """
        try:
            if 0 < header.wfmtype <= 3:  # Vector  # noqa: PLR2004
                waveform = AnalogWaveform()
                waveform.source_name = header.sourcename
                waveform.y_axis_spacing = header.verticalspacing
                waveform.y_axis_offset = header.verticaloffset
                waveform.y_axis_units = header.verticalunits
                waveform.x_axis_spacing = header.horizontalspacing
                waveform.x_axis_units = header.horizontalUnits
                waveform.trigger_index = header.horizontalzeroindex

                sum_of_chunks = 0
                data_size = header.sourcewidth
                request = WaveformRequest(sourcename=header.sourcename, chunksize=self.chunksize)
                response_iterator = self.native.GetWaveform(request)
                dt = None
                sum_chunk_size = 0
                dt_type = self.v_datatypes[header.sourcewidth]

                waveform.y_axis_values = np.empty(header.noofsamples, dtype=dt_type)
                for response in response_iterator:
                    if not self.thread_active:
                        return waveform
                    chunk_size = len(response.headerordata.chunk.data)
                    sum_chunk_size += chunk_size
                    dt = np.frombuffer(response.headerordata.chunk.data, dtype=dt_type)
                    waveform.y_axis_values[
                        sum_of_chunks : sum_of_chunks + int(chunk_size / data_size)
                    ] = dt
                    if dt is not None:
                        sum_of_chunks += len(dt)

            elif header.wfmtype in {7, 6}:  # WFMTYPE_ANALOG_IQ
                waveform = IQWaveform()
                waveform.source_name = header.sourcename
                waveform.iq_axis_spacing = header.verticalspacing
                waveform.iq_axis_offset = header.verticaloffset
                waveform.iq_axis_units = header.verticalunits
                waveform.x_axis_spacing = header.horizontalspacing
                waveform.x_axis_units = header.horizontalUnits
                waveform.trigger_index = header.horizontalzeroindex

                if header.iq_windowType == "Blackharris":
                    sample_rate = (header.iq_fftLength * header.iq_rbw) / 1.9
                elif header.iq_windowType == "Flattop2":
                    sample_rate = (header.iq_fftLength * header.iq_rbw) / 3.77
                elif header.iq_windowType == "Hanning":
                    sample_rate = (header.iq_fftLength * header.iq_rbw) / 1.44
                elif header.iq_windowType == "Hamming":
                    sample_rate = (header.iq_fftLength * header.iq_rbw) / 1.3
                elif header.iq_windowType == "Rectangle":
                    sample_rate = (header.iq_fftLength * header.iq_rbw) / 0.89
                elif header.iq_windowType == "Kaiserbessel":
                    sample_rate = (header.iq_fftLength * header.iq_rbw) / 2.23
                else:
                    sample_rate = header.iq_span

                waveform.meta_info = IQWaveformMetaInfo(
                    iq_center_frequency=header.iq_centerFrequency,
                    iq_fft_length=header.iq_fftLength,
                    iq_resolution_bandwidth=header.iq_rbw,
                    iq_span=header.iq_span,
                    iq_window_type=header.iq_windowType,
                    iq_sample_rate=sample_rate,
                )

                sample_index = 0
                request = WaveformRequest(sourcename=header.sourcename, chunksize=self.chunksize)
                response_iterator = self.native.GetWaveform(request)
                dt = None
                sum_chunk_size = 0
                dt_type = self.iq_datatypes[header.sourcewidth]

                waveform.interleaved_iq_axis_values = np.empty(header.noofsamples, dtype=dt_type)
                for response in response_iterator:
                    if not self.thread_active:
                        return waveform

                    chunk_size = len(response.headerordata.chunk.data)
                    sum_chunk_size += chunk_size
                    dt = np.frombuffer(response.headerordata.chunk.data, dtype=dt_type)
                    sample_count = len(dt)
                    waveform.interleaved_iq_axis_values[
                        sample_index : sample_index + sample_count
                    ] = dt
                    if dt is not None:
                        sample_index += sample_count
            elif header.wfmtype in {4, 5}:  # Digital
                waveform = DigitalWaveform()
                waveform.source_name = header.sourcename
                waveform.y_axis_units = header.verticalunits
                waveform.x_axis_spacing = header.horizontalspacing
                waveform.x_axis_units = header.horizontalUnits
                waveform.trigger_index = header.horizontalzeroindex

                sum_of_chunks = 0
                data_size = header.sourcewidth
                request = WaveformRequest(sourcename=header.sourcename, chunksize=self.chunksize)
                response_iterator = self.native.GetWaveform(request)
                dt = None
                sum_chunk_size = 0
                dt_type = self.d_datatypes[header.sourcewidth]

                waveform.y_axis_byte_values = np.empty(header.noofsamples, dtype=dt_type)
                for response in response_iterator:
                    if not self.thread_active:
                        return waveform
                    chunk_size = len(response.headerordata.chunk.data)
                    sum_chunk_size += chunk_size
                    dt = np.frombuffer(response.headerordata.chunk.data, dtype=dt_type)
                    waveform.y_axis_byte_values[
                        sum_of_chunks : sum_of_chunks + int(chunk_size / data_size)
                    ] = dt
                    if dt is not None:
                        sum_of_chunks += len(dt)

        except Exception as e:
            if self.verbose:
                print_with_timestamp(f"Exception: {e}")

        return waveform

    def _read_waveforms(self, headers, waveforms):
        """Reads the waveforms for the headers.

        Args:
            headers (list): list of headers
            waveforms (list): list of waveforms
        """
        n = len(headers)
        datasize = 0
        for index in range(n):
            header = headers[index]
            read_start = time.perf_counter()
            waveform = self._read_waveform(header)
            self._recordlength = waveform.record_length
            datasize += waveform.record_length * header.sourcewidth
            # TODO: reuse this variable later
            _ = (
                waveform.record_length
                * header.sourcewidth
                * 8
                / ((time.perf_counter() - read_start) * 1e6)
            )
            if self._cachedataenabled:
                self._lock_getdata.acquire()
                self._datacache[header.sourcename.lower()] = waveform
                self._lock_getdata.release()
            if self._recordlength > 0:
                waveforms.append(waveform)
        return datasize

    def _run(self):
        """Background thread for participating in the instruments sequence."""
        while self.thread_active and not self._is_exiting:
            waveforms = []
            headers = []

            startwait = time.perf_counter()
            self._wait_for_data_access()
            self._holding_scope_open = True
            self._lock_filter.acquire()

            try:
                self._run_inner(headers, waveforms, startwait)
            finally:
                self._finished_with_data_access()
                self._lock_filter.release()
                self._holding_scope_open = False

    def _run_inner(self, headers, waveforms, startwait):  # noqa: C901,PLR0912
        """Background thread for participating in the instruments sequence.

        Args:
            headers (list): list of headers
            waveforms (list): list of waveforms
            startwait (float): start time
        """
        datasize = 0
        datawidth = 1
        start = time.perf_counter()
        header_dict = {}

        try:
            if self._cachedataenabled:
                self._lock.acquire(blocking=True)

            if self._is_exiting:
                return

            self._read_headers(headers, header_dict)

            cur_acq_id = self._acq_id(headers)

            if cur_acq_id is None or self._prev_data_id == cur_acq_id:
                return

            self._prev_data_id = cur_acq_id

            if self._filter is not None and not self._filter(self._headers, header_dict):
                self._headers = header_dict
                return

            if len(headers) > 0:
                datawidth = headers[0].sourcewidth

            if self._is_exiting:
                return

            self._headers = header_dict
            datasize += self._read_waveforms(headers, waveforms)
            duration = time.perf_counter() - start
        except Exception as ex:
            if self.verbose:
                print_with_timestamp(f"exception:_run_inner:{ex}")
            # We're exiting so silence any issues and not
            # accumulate bad stats or send bad data
            return

        finally:
            if self._cachedataenabled:
                self._acqtime = self.current_time
                self._lock.release()

        try:
            if len(waveforms) > 0 and self._connected and not self._is_exiting:
                self.data_arrival(waveforms)
                if self._callback is not None:
                    self._callback(waveforms)

        except Exception as ex:
            if self.verbose:
                print_with_timestamp(f"exception:_run_inner:{ex}")

        if self._connected and not self._is_exiting:
            self._acqcount += 1
            self._instrumentation(time.perf_counter() - startwait, duration, datasize, datawidth)

    def _wait_for_acq_time(self, after: float) -> None:
        """Waits until both a new acquisition has arrived and it is later than after.

        Args:
            after (float): Acquisition must occur after this time
        """
        while len(self._datacache) <= 0 or after > self._acqtime:
            self._wait_next_acq()
            if len(self._datacache) <= 0 or after > self._acqtime:
                self._done_with_data_release_lock()
                time.sleep(0.0001)
        self._wait_for_data_count += 1

    def _wait_for_any_acq(self) -> None:
        """Waits for any data to arrive.

        This does not guarantee the data returned is a new acquisition
        """
        while self._acqcount <= 0 or len(self._datacache) <= 0:
            self._wait_next_acq()
            if self._acqcount <= 0 or len(self._datacache) <= 0:
                self._done_with_data_release_lock()
                time.sleep(0.0001)
        self._wait_for_data_count += 1

    def _wait_for_data_access(self) -> None:
        """Waits for instrument server to give the gRPC service a
        chance at the datastore.
        """
        self._in_wait_for_data = True

        if self.verbose:
            print("wait_for_data_access")

        request = ConnectRequest(name=self.clientname)
        self.connection.WaitForDataAccess(request)

    def _wait_for_new_data(self) -> None:
        """Waits for either data from a new acquisition or returns if there.

        is previously unseen data.
        """
        if len(self._datacache) > 0 and self._lastacqseen < self._acqcount:
            self._lock.acquire(blocking=True)
            self._wait_for_data_holds_lock = True
            if self._wait_for_data_count <= 0:
                self._wait_for_data_count = 1
        else:
            self._wait_for_next_acq()

    def _wait_for_next_acq(self) -> None:
        """Waits for the next, new acquisition to arrive."""
        while len(self._datacache) <= 0 or self._lastacqseen >= self._acqcount:
            self._wait_next_acq()
            if len(self._datacache) <= 0 or self._lastacqseen >= self._acqcount:
                self._done_with_data_release_lock()
                time.sleep(0.0001)
        self._wait_for_data_count += 1

    def _wait_next_acq(self):
        self._lock.acquire(blocking=True)
        self._wait_for_data_holds_lock = True

    ################################################################################################
    # Register Methods
    ################################################################################################
    @staticmethod
    @register
    def _terminate():
        """Cleans up mess on termination if possible - this is required
        to keep the scope from hanging
        """
        for key in TekHSIConnect._connections:  # pylint:disable=consider-using-dict-items
            with contextlib.suppress(Exception):
                if TekHSIConnect._connections[key]._holding_scope_open:
                    TekHSIConnect._connections[key]._finished_with_data_access()
            with contextlib.suppress(Exception):
                TekHSIConnect._connections[key].close()



"""Helpers used by the `TekHSI` package."""

from tekhsi.helpers.constants import PACKAGE_NAME
from tekhsi.helpers.functions import print_with_timestamp

__all__ = [
    "PACKAGE_NAME",
    "print_with_timestamp",
]
